#include <iostream>
#include <stdexcept>

#include <Eigen/Dense>

#ifdef CSDP_SOLVER
#include <gposolver/gposolver_csdp.h>
#elif MOSEK_SOLVER
#include <gposolver/gposolver_mosek.h>
#elif SDPA_SOLVER
#include <gposolver/gposolver_sdpa.h>
#else
#error Define CSDP_SOLVER, MOSEK_SOLVER, or SDPA_SOLVER macros
#endif

#include "hec_problem.h"

using namespace std;
using namespace Eigen;
using namespace GpoSolver;


void q2rot(Vector4d &q, Matrix3d &R) {
R << q(0)*q(0) + q(1)*q(1) - q(2)*q(2) - q(3)*q(3),                     2*q(1)*q(2) - 2*q(0)*q(3),                     2*q(1)*q(3) + 2*q(0)*q(2),
                         2*q(1)*q(2) + 2*q(0)*q(3), q(0)*q(0) - q(1)*q(1) + q(2)*q(2) - q(3)*q(3),                     2*q(2)*q(3) - 2*q(0)*q(1), 
                         2*q(1)*q(3) - 2*q(0)*q(2),                     2*q(2)*q(3) + 2*q(0)*q(1), q(0)*q(0) - q(1)*q(1) - q(2)*q(2) + q(3)*q(3);
}

int main(int argc, const char* argv[]) {
#ifdef CSDP_SOLVER
  GpoSolverCsdp<HecProblem> gposolver;
#elif MOSEK_SOLVER
  GpoSolverMosek<HecProblem> gposolver;
#elif SDPA_SOLVER
  GpoSolverSdpa<HecProblem> gposolver;
#endif

  try
    {
      GpoSolverStatus status;
      GpoSolverSolutions sols;
      MatrixXd rvals;
      VectorXd pvals;

      if (argc != 2)
        {
          cout << "Usage: hec_solver hec_data.txt" << endl;
          return 0;
        }

      // Load problem data using GpoSolver utility function 
      gposolver.readProblemDataTxt(argv[1], rvals, pvals);

      // Scale translations
      double max_norm = 1;
      for (int i = 0; i < rvals.cols(); i++)
        {
          double ta_norm = rvals.block(9, i, 3, 1).norm();
          double tb_norm = rvals.block(21, i, 3, 1).norm();

          if (ta_norm > max_norm)
            max_norm = ta_norm;

          if (tb_norm > max_norm)
            max_norm = tb_norm;
        }

      for (int i = 0; i < rvals.cols(); i++)
        {
          rvals.block(9, i, 3, 1) /= max_norm;
          rvals.block(21, i, 3, 1) /= max_norm;
        }

      // Call GpoSolver
      gposolver.setParameter("verbose", 1);
      status = gposolver.solve(rvals.cols(), rvals.data(), pvals.data(), sols);

      // Print solution
      cout << endl << endl << "GpoSolver: " << sols.size() << " solution(s) extracted" << endl;

      if (sols.size() == 0)
        return 0;

      if (status == SUCCESS_GLOPT)
        cout << "Global optimality certified numerically" << endl;
      else if (status == SUCCESS)
        cout << "Alas, global optimality could not be certified" << endl;

      if (sols.size() > 0)
        {
          int c = 0;
          for (list<vector<double> >::iterator it = sols.begin(); it != sols.end(); it++)
            {
              Matrix4d X;
              Matrix3d R;
              Vector4d q;
              Vector3d t;

              for (int i = 0; i < 4; i++)
                q(i) = (*it)[i];
              for (int i = 0; i < 3; i++)
                t(i) = (*it)[i + 4];

              q2rot(q, R);
              X.setIdentity();
              X.block(0, 0, 3, 3) = R;
              X.block(0, 3, 3, 1) = t * max_norm;
		
              cout << endl << "X(" << ++c << ") = " << endl << X << endl;
            }
        }
    }
  catch (exception& e)
    {
      cerr << "Solver exception: " << e.what() << endl;
      return 1;
    }

  return 0;
}
