#include <iostream>
#include <stdexcept>

#include <Eigen/Dense>

#ifdef CSDP_SOLVER
#include <gposolver/gposolver_csdp.h>
#elif MOSEK_SOLVER
#include <gposolver/gposolver_mosek.h>
#elif SDPA_SOLVER
#include <gposolver/gposolver_sdpa.h>
#else
#error Define CSDP_SOLVER, MOSEK_SOLVER, or SDPA_SOLVER macros
#endif

#include "pnp_problem.h"

using namespace std;
using namespace Eigen;
using namespace GpoSolver;


void q2rot(Vector4d &q, Matrix3d &R) {
R << q(0)*q(0) + q(1)*q(1) - q(2)*q(2) - q(3)*q(3),                     2*q(1)*q(2) - 2*q(0)*q(3),                     2*q(1)*q(3) + 2*q(0)*q(2),
                         2*q(1)*q(2) + 2*q(0)*q(3), q(0)*q(0) - q(1)*q(1) + q(2)*q(2) - q(3)*q(3),                     2*q(2)*q(3) - 2*q(0)*q(1), 
                         2*q(1)*q(3) - 2*q(0)*q(2),                     2*q(2)*q(3) + 2*q(0)*q(1), q(0)*q(0) - q(1)*q(1) - q(2)*q(2) + q(3)*q(3);
}

int main(int argc, const char* argv[]) {
#ifdef CSDP_SOLVER
  GpoSolverCsdp<PnpProblem> gposolver;
#elif MOSEK_SOLVER
  GpoSolverMosek<PnpProblem> gposolver;
#elif SDPA_SOLVER
  GpoSolverSdpa<PnpProblem> gposolver;
#endif
  
  gposolver.setParameter("restol", 0.1);

  try
    {
      GpoSolverStatus status;
      GpoSolverSolutions sols;
      MatrixXd rvals;
      VectorXd pvals;

      if (argc != 2)
        {
          cout << "Usage: pnp_solver pnp_data.txt" << endl;
          return 0;
        }

      // Load problem data using GpoSolver utility function 
      gposolver.readProblemDataTxt(argv[1], rvals, pvals);

      // Compute T matrix
      MatrixXd T(3, 10);
      MatrixXd Qs(3, 3);
      MatrixXd Cs(3, 10);
      Qs.setZero();
      Cs.setZero();

      for (int i = 0; i < rvals.cols(); i++)
        {
	  Vector3d v = rvals.block(0, i, 3, 1);
	  Vector3d c = rvals.block(3, i, 3, 1);
	  Vector3d x = rvals.block(6, i, 3, 1);

          // Normalize view direction
	  v.normalize();
	  rvals.block(0, i, 3, 1) = v;

	  Matrix3d Q = Matrix3d::Identity() - v * v.transpose();
          MatrixXd C(3, 10);

          C << x(0), x(1), x(2), 0, 0, 0, 0, 0, 0, c(0),
               0, 0, 0, x(0), x(1), x(2), 0, 0, 0, c(1),
               0, 0, 0, 0, 0, 0, x(0), x(1), x(2), c(2);

          Qs += Q;
          Cs += Q * C;            
        }

      T = -1 * (Qs.inverse() * Cs);
      pvals = Map<VectorXd>(T.data(), 3 * 10);

      // Call GpoSolver
      gposolver.setParameter("verbose", 1);
      status = gposolver.solve(rvals.cols(), rvals.data(), pvals.data(), sols);

      // Print solution
      cout << endl << endl << "GpoSolver: " << sols.size() << " solution(s) extracted" << endl;

      if (sols.size() == 0)
        return 0;

      if (status == SUCCESS_GLOPT)
        cout << "Global optimality certified numerically" << endl;
      else if (status == SUCCESS)
        cout << "Alas, global optimality could not be certified" << endl;

      if (sols.size() > 0)
        {
          int c = 0;
          for (list<vector<double> >::iterator it = sols.begin(); it != sols.end(); it++)
            {
              Matrix4d A;
              Matrix3d R;
              Vector4d q;
              VectorXd r(10);

              for (int i = 0; i < 4; i++)
                q(i) = (*it)[i];

	      // Make q a perfect unit quaternion
	      q.normalize();
              q2rot(q, R);

	      r.head(9) =  Map<VectorXd>(R.data(), 3 * 3);
	      r(9) = 1;
 	
              A.setIdentity();
              A.block(0, 0, 3, 3) = R;
              A.block(0, 3, 3, 1) = T * r;
		
              cout << endl << "A(" << ++c << ") = " << endl << A << endl;
            }
        }
    }
  catch (exception& e)
    {
      cerr << "Solver exception: " << e.what() << endl;
      return 1;
    }

  return 0;
}
